import {Injectable} from '@angular/core';
import {HttpInterceptor, HttpRequest, HttpHandler, HttpEvent, HttpErrorResponse, HTTP_INTERCEPTORS, HttpResponse} from '@angular/common/http';
import { Observable, throwError } from 'rxjs';
import { catchError } from 'rxjs/operators';
import { ToasterWrapperService } from '../_services/toaster-wrapper.service';
import { AuthenticationService } from '../_services/authentication.service';
import { Router } from '@angular/router';
import { resolve } from 'dns';
import { TranslatorService } from '../_services/translator.service';
import { TranslateService } from '@ngx-translate/core';
import { ExclusionHeaders } from './exclusion-headers';


@Injectable()
export class ErrorInterceptor implements HttpInterceptor {

    constructor(private toasterService: ToasterWrapperService,
                private translateService: TranslateService) { }

    intercept(req: HttpRequest<any>, next: HttpHandler): Observable<HttpEvent<any>> {
        
        if (req.headers.has(ExclusionHeaders.ERROR_INTERCEPTOR)){
            return next.handle(req);
        }

        return next.handle(req).pipe(
            catchError(error => {
                if (error instanceof HttpErrorResponse) {
                    if (error.status === 401) {
                        this.toasterService.toastError(error.statusText);

                        return throwError(error.statusText);
                    } else if (error.status === 0) {
                        const noBackendConnectionError = this.translateService.instant('misc.NO_CONNESSIONE_BACKEND');

                        this.toasterService.toastError(noBackendConnectionError);
                        return throwError(noBackendConnectionError);
                    } else if (error instanceof HttpResponse) {
                        const applicationError = (error as HttpResponse<string>).body;

                        this.toasterService.toastError(applicationError);

                        return throwError(applicationError);
                    }

                    const composedError = this.getComposedError(error);

                    this.toasterService.toastError(composedError);

                    return throwError(composedError);
                }

                return throwError(error);
            })
        );
    }

    private getComposedError(error){
        const serverError = error.error;
        let modalStateErrors = '';
        if (serverError && typeof serverError === 'object') {
            for (const key in serverError) {
                if (serverError[key]) {
                    modalStateErrors += serverError[key] + '\n';
                }
            }
        }

        return modalStateErrors || serverError || 'Server Error';
    }
}

export const ErrorInterceptorProvider = {
    provide: HTTP_INTERCEPTORS,
    useClass: ErrorInterceptor,
    multi: true
};

// estrazione errore direttamente da header -> problema: CODIFICA SOLO ASCII!!!
/* const applicationError = error.headers.get('Application-Error');
    if (applicationError) {
        this.toasterService.toastError(applicationError);

        return throwError(applicationError);
    } */