import { Injectable } from '@angular/core';
import { HttpClient } from '@angular/common/http';
import { BehaviorSubject, Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import { environment } from 'src/environments/environment';
import { AuthResult } from '../_models/auth-result';
import { Router } from '@angular/router';
import jwt_decode from 'jwt-decode';
import { AuthRequest } from '../_models/auth-request';
import { ChatSignalRService } from './chat-signalr.service';
import { SettingsService } from './settings.service';

// da tenere allineato con UserPermissions nel backend
export class UserPermissions {
    public static Permesso1 = 'PERMESSO_1';
    public static Permesso1ID = 0;

    public static Permesso2 = 'PERMESSO_2';
    public static Permesso2ID = 1;
}

// da tenere allineato con UserLevels nel backend
export class UserLevels {
    public static Default = 'DEFAULT';
}

// da tenere allineato con UserAccesses nel backend
export class UserAccesses {
    public static Admin = 'ADMIN';
    public static Standard = 'STANDARD';
}

@Injectable()
export class AuthenticationService {
    private currentUserSubject: BehaviorSubject<AuthResult>;
    public currentUser: Observable<AuthResult>;
    private tokenExpirationTimer: any;
    private autoLogoutTimer: any;

    constructor(private http: HttpClient, private router: Router,
        private chatServ: ChatSignalRService, private settingsServ: SettingsService) {
        this.currentUserSubject = new BehaviorSubject<AuthResult>(null);

        // chiamerà il next su "currentUserSubject" passando NULL
        // o l'utente stesso (se presente in localStorage e non scaduto)
        // e si occupa di gestire il timer per l'autologout
        this._initCurrentUser(JSON.parse(localStorage.getItem('currentUser')));

        this.currentUser = this.currentUserSubject.asObservable();

        document.onmousemove = () => {
            if (this.currentUserSubject && this.currentUserSubject.value) {
                this._resetAutoLogoutTimer();
            }
        };
        document.onclick = () => {
            if (this.currentUserSubject && this.currentUserSubject.value) {
                this._resetAutoLogoutTimer();
            }
        };
        document.onkeypress = () => {
            if (this.currentUserSubject && this.currentUserSubject.value) {
                this._resetAutoLogoutTimer();
            }
        };
    }

    public get currentUserValue(): AuthResult {
        return this.currentUserSubject.value;
    }

    public get currentUserToken(): string{
        if (this.currentUserValue){
            return this.currentUserValue.Token;
        }
        return null;
    }

    private _initCurrentUser(user: AuthResult) {
        // se nessun utente è stato passato..
        if (!user) {
            // rimuovo l'utente
            this._clearUser();
            return;
        }

        const decodedToken = jwt_decode(user.Token);
        const expirationDuration = this._getTokenRemainingDurationMsec(decodedToken);

        // se la durata residua è minore o uguale a 0 (per cui il token è scaduto..)
        if (expirationDuration <= 0) {
            // rimuovo l'utente
            this._clearUser();
            return;
        }

        // altrimenti, l'utente è valido e il token non è scaduto:
        // pertanto, lo imposto come utente e creo il timer per fare
        // logout in automatico allo scadere del token
        this._setUser(user, expirationDuration);

        // avvio il servizio di chat
        this.chatServ.start(user.Utente);
    }

    private _setUser(user, expirationDuration){
        this.tokenExpirationTimer = setTimeout(() => this.logout(), expirationDuration);
        this._resetAutoLogoutTimer();
        this.currentUserSubject.next(user);
    }

    private _resetAutoLogoutTimer() {
        if (this.autoLogoutTimer) {
            clearTimeout(this.autoLogoutTimer);
        }
        const autoLogoutMinutes = this.settingsServ.getAppSetting('autoLogoutMinuti');
        this.autoLogoutTimer = setTimeout(() => this.logout(), autoLogoutMinutes * 60000);
    }

    private _clearUser(){
        // rimuovo l'eventuale timer per
        // l'autologout allo scadere del token
        if (this.tokenExpirationTimer) {
            clearTimeout(this.tokenExpirationTimer);
        }

        // aggiorno l'utente corrente con il valore NULL
        // (qualora precedentemente fosse stato valorizzato..)
        this.currentUserSubject.next(null);
    }

    private _getTokenRemainingDurationMsec(token){
        // "exp" sta per "Expiry"
        // "iat" sta per "Issued at" (data di emissione, non utilizzato)
        // mentre "nbf" sta per "Not Before" (non utilizzato)
        return (new Date(token.exp * 1000).getTime() - new Date().getTime());
    }

    login(credentials: AuthRequest) {
        const loginUrl: string = environment.apiUrl + '/users/authenticate';

        return this.http.post<AuthResult>(loginUrl, credentials)
            .pipe(map(user => {
                // login successful if there's a jwt token in the response
                if (user && user.Token) {

                    // salvo l'utente corrente (che contiene anche il token) 
                    // nel localStorage..
                    localStorage.setItem('currentUser', JSON.stringify(user));

                    // ..e notifico all'applicazione che l'utente è cambiato
                    this._initCurrentUser(user);
                }

                return user;
            }));
    }

    logout() {
        // remove user from local storage to log user out
        localStorage.removeItem('currentUser');
        this.currentUserSubject.next(null);

        // stoppo il servizio di chat
        this.chatServ.stop();

        this.router.navigate(['./login']);
    }

    isAdmin(user: AuthResult): boolean {
        if (!this.currentUserValue) {
            return false;
        }

        return user.Roles
            && user.Roles.includes(UserAccesses.Admin);
    }

    isIdUtenteLoggato(userId: number) {
        if (!this.currentUserValue) {
            return false;
        }

        return this.currentUserValue.Utente.Id === userId;
    }

    isLivelloDipendente(userLevelRole: string){
        if (!this.currentUserValue){
            return false;
        }

        return this.currentUserValue.Roles
            && this.currentUserValue.Roles.includes(userLevelRole);
    }

    isAuthorized(allowedRoles: string[]): boolean {
        // se allowedRoles è undefined l'accesso è consentito anche senza fare login
        if (!allowedRoles) {
            return true;
        }

        // altrimenti, se l'utente è admin, posso fare tutto
        if (this.isAdmin(this.currentUserValue)) {
            return true;
        }

        // se arrivo qui l'utente non è admin: 
        // se non è neanche loggato NON posso procedere
        if (!this.currentUserValue) {
            return false;
        }

        // se l'utente fra i suoi ruoli ha uno di quelli previsti, posso procedere
        for (const role of allowedRoles) {
            if (this.currentUserValue.Roles.includes(role)) {
                return true;
            }
        }

        // altrimenti no
        return false;
      }
}
