import { Injectable } from '@angular/core';
import { environment } from 'src/environments/environment';
import { BehaviorSubject } from 'rxjs';
import { Message, UserChatData } from '../_models/chat';
import { Utente } from '../_models/common';
import { ToasterWrapperService } from './toaster-wrapper.service';
import * as signalR from '@microsoft/signalr';

@Injectable()
export class ChatSignalRService {

constructor(private toastServ: ToasterWrapperService) { }

  private _isChatOpen = false;
  private _userChatData = 'user-chat-data';
  private _maxHistoryCount = 200;

  private lastSeen: Date;

  public $isConnected = new BehaviorSubject<boolean>(false);
  public $messages: BehaviorSubject<Message[]>;
  public $newMessagesCount: BehaviorSubject<number>;

  private _hubConnection: signalR.HubConnection;

  private _user : Utente;

  // in fase di apertura iniziale della connessione, se l'operazione
  // fallisce, si riprova dopo il tempo indicato (in ms)
  private _startConnectionRetryDelay = 5000;
  // se la connessione è caduta e si sta cercando di ripristinarla,
  // fino a che non è passato un tempo >= a "_reconnectThresholdDelay"
  // l'applicazione prova continuamente a riconnettersi ogni t_random,
  // con t_random incluso nell'intervallo [0, "_reconnectBelowThresholdMaxDelay"],
  // oltre tale threshold, l'applicazione riproverà a riconnettersi ogni
  // "_reconnectBelowThresholdMaxDelay" (in ms)
  private _reconnectBelowThresholdMaxDelay = 10000;
  private _reconnectThresholdDelay = 60000;

  public get isOpen(){
    return this._isChatOpen;
  }

  public openChat(){
    if (!this.$isConnected.value){
      this.toastServ.toastError('Al momento il servizio di chat non è disponibile. Riprovare in un secondo momento');
      return;
    }

    console.log('chat window opening...')
    this._isChatOpen = true;

    // ottengo un riferimento a tutti e soli i messaggi
    // presenti nel momento in cui la chat viene aperta..
    // (altri messaggi potrebbero essere inviati dopo..)
    const messagesOnOpen = [...this.$messages.value];

    setTimeout(() => {
      // se nel frattempo la chat è stata chiusa, non faccio nulla..
      if (!this._isChatOpen){
        return;
      }

      const now = new Date();

      if (!this.lastSeen || now > this.lastSeen){
        this.lastSeen = now;
      }

      // aggiorno lo stato di lastSeen su localStorage
      // perchè questo metodo potrebbe averlo aggiornato
      this._setUserChatDataToLocalStorage(this.lastSeen);

      for(const message of messagesOnOpen) {
        message.IsNew = false;
      }

      // refresh messaggi e conteggio: qui utilizzo il valore corrente di $messages
      // perchè nel frattempo potrebbero essersi aggiunti altri messaggi
      this.$messages.next(this.$messages.value);
      this.$newMessagesCount.next(this.$messages.value.filter(m => m.IsNew).length);
    }, 5000);
  }

  public closeChat() {
    console.log('chat window closing...');
    this._isChatOpen = false;
  }

  public sendMessage(message): void {
    this._hubConnection
      .invoke('broadcastMessage', this._user.IdentificativoUtente, message)
      // .then(() => this.message = '')
      .catch(err => console.error(err));
  }

  public async stop() {
    if (!this._hubConnection){
        return;
    }

    await this._hubConnection.stop();

    this._user = null;
    this.$isConnected.next(false);
    console.log('Chat SignalrR connection closed');
  }

  public async start(utente: Utente) {
    this._user = utente;

    this._hubConnection = new signalR.HubConnectionBuilder()
      .withUrl(environment.apiUrl.replace('/api', '') + '/chat')
      .withAutomaticReconnect({ nextRetryDelayInMilliseconds: retryContext => this._nextRetryDelayInMilliseconds(retryContext) })
      .build();

    this._hubConnection.onclose((error) => {
      // in caso di disconnessione, chiudo la chat, indico ai
      // listener esterni che la connessione è chiusa, e poi
      // lancio il metodo che ogni tot secondi prova ad
      // effettuare la riconnessione
      this.closeChat();
      this.$isConnected.next(false);
      console.log('disconnected from signalR: ' + error);
    });

    this._hubConnection.on('receiveMessagesHistory', (sortedHistoryMessages) => {
      // fix date da stringa a date (TODO metodo più sensato?)
      for(const message of sortedHistoryMessages){
        message.Date = new Date(message.Date);
      }

      const userChatData = this._getUserChatDataFromLocalStorage();

        if (userChatData.LastSeen){
          this.lastSeen = userChatData.LastSeen;
        }

        this._setMessagesCustomProps(sortedHistoryMessages);
        this.$messages = new BehaviorSubject<Message[]>(sortedHistoryMessages);

        const newMessagesCount = sortedHistoryMessages.filter(m => m.IsNew).length;
        this.$newMessagesCount = new BehaviorSubject<number>(newMessagesCount);
      });

    this._hubConnection.on('receiveMessage', (newMessage) => {
      // fix date da stringa a date (TODO metodo più sensato?)
      newMessage.Date = new Date(newMessage.Date);

      if (this._isChatOpen){
        this.lastSeen = newMessage.Date;
      }

      const updatedMessages = this._getUpdatedMessages(newMessage);
      this._setMessagesCustomProps(updatedMessages);

      this._setUserChatDataToLocalStorage(this.lastSeen);

      this.$messages.next(updatedMessages);
      this.$newMessagesCount.next(updatedMessages.filter(m => m.IsNew).length);
    });

    await this._startConnection();
  }

  private async _startConnection(){
    try {
      await this._hubConnection.start();
      console.assert(this._hubConnection.state === signalR.HubConnectionState.Connected);
      this.$isConnected.next(true);
      console.log('Chat SignalR connection open');
    } catch (err) {
      console.assert(this._hubConnection.state === signalR.HubConnectionState.Disconnected);
      this.$isConnected.next(false);
      console.log('Error while establishing SignalR chat connection: ' + err);
      setTimeout(() => this._startConnection(), this._startConnectionRetryDelay);
    }
  }

  private _nextRetryDelayInMilliseconds(retryContext){
    if (retryContext.elapsedMilliseconds < this._reconnectThresholdDelay) {
        return Math.random() * this._reconnectBelowThresholdMaxDelay;
    } else {
        // If we've been reconnecting for more than 60 seconds so far, stop reconnecting.
        return this._reconnectThresholdDelay;
    }
  }

  private _setMessagesCustomProps(sortedHistoryMessages: Message[]) {
    for(const message of sortedHistoryMessages) {
      message.IsNew = (!this.lastSeen || message.Date > this.lastSeen) ? true : false;
      message.IsOwn = message.User === this._user.IdentificativoUtente;
    }
  }

  private _getUpdatedMessages(newMessage){
    const toReturn = this.$messages.value;
    toReturn.push(newMessage);

    while (toReturn.length > this._maxHistoryCount){
      toReturn.splice(0, 1);
    }

    return toReturn;
  }

  private _setUserChatDataToLocalStorage(lastSeen:Date) {
    const updatedChatData = new UserChatData();
    updatedChatData.LastSeen = lastSeen;

    localStorage.setItem(this._userChatData, JSON.stringify(updatedChatData));
  }

  private _getUserChatDataFromLocalStorage(): UserChatData {
      const JSONToReturn = localStorage.getItem(this._userChatData);

      const toReturn = (!JSONToReturn)
      ? new UserChatData()
      : JSON.parse(JSONToReturn);

      if (toReturn.LastSeen){
        toReturn.LastSeen = new Date(toReturn.LastSeen);
      }

      return toReturn;
  }
}
