import { Injectable, Type } from '@angular/core';
import { DBConfig, Key, NgxIndexedDBService } from 'ngx-indexed-db';
import { forkJoin, from, Observable, of, throwError } from 'rxjs';
import { concatMap, defaultIfEmpty, map, switchMap, take, takeLast, tap } from 'rxjs/operators';
import { v4 as uuidv4 } from 'uuid';
import { AnomaliaDB, BooleanState, ChecklistDB, ComponenteDB, ElementoChecklistDB, FullChecklistDB, FullSessioneDB, DetailedStoricoDifetto, FullVeicoloDB, GammaDB, GammaSpecificaDB, GruppoDB, ParticolareDB, PesoDifettoDB, PosizioneAnomaliaDB, ProdottoDB, ResponsabilitaDB, SessioneDB, StoricoDifettoDB, VeicoloDB, TipoSessione } from '../_models/common';
import { TranslatorService } from './translator.service';
import { UtilsService } from './utils.service';

const prodottiStore = 'prodotti';
const gruppiStore = 'gruppi';
const particolariStore = 'particolari';
const componentiStore = 'componenti';
const anomalieStore = 'anomalie';
const posAnomalieStore = 'posizioniAnomalie';
const pesiDifettoStore = 'pesiDifetto';
const responsabilitaStore = 'responsabilita';
const gammeStore = 'gamme';
const gammeSpecificheStore = 'gammeSpecifiche';
const checklistsStore = 'checklists';
const elementiChecklistsStore = 'elementiChecklists';

const veicoliStore = 'veicoli';
const sessioniStore = 'sessioni';
const storicoDifettiStore = 'storicoDifetti';

export const dbConfig: DBConfig  = {
    name: 'frontend-db',
    version: 1,
    objectStoresMeta: [
    {
      store: prodottiStore,
      storeConfig: { keyPath: 'Codice', autoIncrement: false },
      storeSchema: [
        { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
      ]
    },
    {
        store: gruppiStore,
        storeConfig: { keyPath: 'Codice', autoIncrement: false },
        storeSchema: [
          { name: 'CodiceParent', keypath: 'CodiceParent', options: { unique: false } },
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: particolariStore,
        storeConfig: { keyPath: 'Codice', autoIncrement: false },
        storeSchema: [
          { name: 'CodiceParent', keypath: 'CodiceParent', options: { unique: false } },
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: componentiStore,
        storeConfig: { keyPath: 'Codice', autoIncrement: false },
        storeSchema: [
          { name: 'CodiceParent', keypath: 'CodiceParent', options: { unique: false } },
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: anomalieStore,
        storeConfig: { keyPath: 'Codice', autoIncrement: false },
        storeSchema: [
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } },
          { name: 'Audit', keypath: 'Audit', options: { unique: false } },
        ]
      },
      {
        store: posAnomalieStore,
        storeConfig: { keyPath: 'Codice', autoIncrement: false },
        storeSchema: [
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: pesiDifettoStore,
        storeConfig: { keyPath: 'Id', autoIncrement: false },
        storeSchema: [
          { name: 'TipoDifetto', keypath: 'TipoDifetto', options: { unique: false } },
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: responsabilitaStore,
        storeConfig: { keyPath: 'Id', autoIncrement: false },
        storeSchema: [
          { name: 'GammaId', keypath: 'GammaId', options: { unique: false } },
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: gammeStore,
        storeConfig: { keyPath: 'Id', autoIncrement: false },
        storeSchema: [
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: gammeSpecificheStore,
        storeConfig: { keyPath: 'Id', autoIncrement: false },
        storeSchema: [
          { name: 'GammaId', keypath: 'GammaId', options: { unique: false } },
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: checklistsStore,
        storeConfig: { keyPath: 'Id', autoIncrement: false },
        storeSchema: [
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: elementiChecklistsStore,
        storeConfig: { keyPath: 'Id', autoIncrement: false },
        storeSchema: [
          { name: 'ChecklistId', keypath: 'ChecklistId', options: { unique: false } },
          { name: 'Abilitato', keypath: 'Abilitato', options: { unique: false } }
        ]
      },
      {
        store: veicoliStore,
        storeConfig: { keyPath: 'GUID', autoIncrement: false },
        storeSchema: [
          { name: 'VAN', keypath: 'VAN', options: { unique: true } },
          { name: 'Categoria', keypath: 'Categoria', options: { unique: false } },
        ]
      },
      {
        store: sessioniStore,
        storeConfig: { keyPath: 'GUID', autoIncrement: false },
        storeSchema: [
          { name: 'VeicoloGUID', keypath: 'VeicoloGUID', options: { unique: false } },
        ]
      },
      {
        store: storicoDifettiStore,
        storeConfig: { keyPath: 'GUID', autoIncrement: false },
        storeSchema: [
          { name: 'SessioneGUID', keypath: 'SessioneGUID', options: { unique: false } },
          { name: 'Codice', keypath: 'Codice', options: { unique: false } },
          { name: 'ProdottoCodice', keypath: 'ProdottoCodice', options: { unique: false } },
          { name: 'GruppoCodice', keypath: 'GruppoCodice', options: { unique: false } },
        ]
      },
  ]
  };

@Injectable()
export class DbWrapperService {

  private readonly _DIFETTO_NO_PARTICOLARE_AFFIX = 'DD';
  private readonly _DIFETTO_NO_COMPONENTE_AFFIX = 'CC';

  constructor(private localDBServ: NgxIndexedDBService, private translServ: TranslatorService, 
    private utilsServ: UtilsService) { }

  getProdotti(onlyEnabled: boolean){
      return this._getEntitiesWithAbilitato<ProdottoDB>(prodottiStore, 'Abilitato', onlyEnabled);
  }

  getProdottoByCodice(codice: string): Observable<ProdottoDB> {
    return this._getEntityOrNull<ProdottoDB>(prodottiStore, codice);
  }

  getGruppi(onlyEnabled: boolean, codiceParent: string = null){
      const toReturn = this._getEntitiesWithParent<GruppoDB, string>(gruppiStore, 'CodiceParent', codiceParent);

      return toReturn.pipe(
          map(result => (onlyEnabled) ? result.filter(m => m.Abilitato === BooleanState.True) : result)
      );
  }

  getGruppoByCodice(codice: string): Observable<GruppoDB> {
    return this._getEntityOrNull<GruppoDB>(gruppiStore, codice);
  }

  getParticolari(onlyEnabled: boolean, codiceParent: string = null){
      const toReturn = this._getEntitiesWithParent<ParticolareDB, string>(particolariStore, 'CodiceParent', codiceParent);

      return toReturn.pipe(
          map(result => (onlyEnabled) ? result.filter(m => m.Abilitato === BooleanState.True) : result)
      );
  }

  getParticolareByCodice(codice: string): Observable<ParticolareDB> {
    return this._getEntityOrNull<ParticolareDB>(particolariStore, codice);
  }

  getComponenti(onlyEnabled: boolean, codiceParent: string = null){
      const toReturn = this._getEntitiesWithParent<ComponenteDB, string>(componentiStore, 'CodiceParent', codiceParent);

      return toReturn.pipe(
          map(result => (onlyEnabled) ? result.filter(m => m.Abilitato === BooleanState.True) : result)
      );
  }

  getComponenteByCodice(codice: string): Observable<ComponenteDB> {
    return this._getEntityOrNull<ComponenteDB>(componentiStore, codice);
  }

  getAnomalie(onlyEnabled: boolean, onlyForAudit: boolean){
    const toReturn = this._getEntitiesWithAbilitato<AnomaliaDB>(anomalieStore, 'Abilitato', onlyEnabled);

    return toReturn.pipe(
        map(result => (onlyForAudit) ? result.filter(m => m.Audit === BooleanState.True) : result)
    );
  }

  getAnomaliaByCodice(codice: string): Observable<AnomaliaDB> {
    return this._getEntityOrNull<AnomaliaDB>(anomalieStore, codice);
  }

  getPosAnomalie(onlyEnabled: boolean){
      return this._getEntitiesWithAbilitato<PosizioneAnomaliaDB>(posAnomalieStore, 'Abilitato', onlyEnabled);
  }

  getPosAnomaliaByCodice(codice: string): Observable<PosizioneAnomaliaDB> {
    return this._getEntityOrNull<PosizioneAnomaliaDB>(posAnomalieStore, codice);
  }

  getPesiDifetto(onlyEnabled: boolean){
      return this._getEntitiesWithAbilitato<PesoDifettoDB>(pesiDifettoStore, 'Abilitato', onlyEnabled);
  }

  getPesoDifettoById(id: number): Observable<PesoDifettoDB> {
    return this._getEntityOrNull<PesoDifettoDB>(pesiDifettoStore, id);
  }

  getResponsabilita(onlyEnabled: boolean){
      return this._getEntitiesWithAbilitato<ResponsabilitaDB>(responsabilitaStore, 'Abilitato', onlyEnabled);
  }

  getGamme(onlyEnabled: boolean){
    return this._getEntitiesWithAbilitato<GammaDB>(gammeStore, 'Abilitato', onlyEnabled);
  }

  getGammaById(id: number): Observable<GammaDB> {
    return this._getEntityOrNull<GammaDB>(gammeStore, id);
  }

  getGammeSpecifiche(onlyEnabled: boolean){
      return this._getEntitiesWithAbilitato<GammaSpecificaDB>(gammeSpecificheStore, 'Abilitato', onlyEnabled);
  }

  getGammaSpecificaById(id: number): Observable<GammaSpecificaDB> {
    return this._getEntityOrNull<GammaSpecificaDB>(gammeSpecificheStore, id);
  }

  getChecklists(onlyEnabled: boolean){
    return this._getEntitiesWithAbilitato<ChecklistDB>(checklistsStore, 'Abilitato', onlyEnabled);
  }

  getChecklistById(id: number): Observable<ChecklistDB> {
    return this._getEntityOrNull<ChecklistDB>(checklistsStore, id);
  }

  getFullChecklists(onlyEnabled: boolean): Observable<FullChecklistDB[]> {
    return this.getChecklists(onlyEnabled).pipe(
      switchMap(m => forkJoin([ ...m.map(m => this._getFullChecklistDB(m)) ])),
      defaultIfEmpty([])
    );
  }

  getFullChecklistById(id: number): Observable<FullChecklistDB> {
    return this.getChecklistById(id).pipe(
      switchMap(m => this._getFullChecklistDB(m))
    );
  }

  getElementiChecklists(onlyEnabled: boolean, idChecklist: number = null){
    const toReturn = this._getEntitiesWithParent<ElementoChecklistDB, number>(elementiChecklistsStore, 'ChecklistId', idChecklist);

    return toReturn.pipe(
        map(result => (onlyEnabled) ? result.filter(m => m.Abilitato === BooleanState.True) : result)
    );
  }

  getDifettoByGUID(GUID: string): Observable<StoricoDifettoDB>{
    return this.localDBServ.getByKey<StoricoDifettoDB>(storicoDifettiStore, GUID);
  }

  getDetailedDifettoByGUID(GUID: string): Observable<DetailedStoricoDifetto> {
    return this.getDifettoByGUID(GUID)
      .pipe(
        switchMap(difetto => this._getDetailedDifetto(difetto))
      );
  }

  getDifettiSessione(sessioneGUID: string){
      return this.localDBServ.getAllByIndex<StoricoDifettoDB>(storicoDifettiStore, 'SessioneGUID', IDBKeyRange.only(sessioneGUID));
  }

  getDetailedDifettiSessione(GUIDsessione: string): Observable<DetailedStoricoDifetto[]> {
    return this.getDifettiSessione(GUIDsessione)
    .pipe(
      switchMap(difetti => forkJoin([ ...difetti.map(m => this._getDetailedDifetto(m))])),
      defaultIfEmpty([])
    );
  }

  deleteStoricoDifetto(difettoGUID: string){
    return this.localDBServ.deleteByKey(storicoDifettiStore, difettoGUID);
  }

  saveOrUpdateStoricoDifetto(toSaveOrUpdate: StoricoDifettoDB){
    if (toSaveOrUpdate.GUID){

      // il codice (non univoco) potrebbe essere cambiato se i dati abbinati sono stati modificati
      toSaveOrUpdate.Codice = this._getCodiceDifetto(toSaveOrUpdate);

      return this._updateEntity(storicoDifettiStore, toSaveOrUpdate, toSaveOrUpdate.GUID);
    } else {
      // in fase di salvataggio iniziale del difetto, attribuisco sia il GUID
      // che il codice (non univoco), in base ai dati abbinati al difetto
      toSaveOrUpdate.GUID = uuidv4();
      toSaveOrUpdate.Codice = this._getCodiceDifetto(toSaveOrUpdate);

      return this._addEntity<StoricoDifettoDB>(storicoDifettiStore, toSaveOrUpdate);
    }
  }

  editDifettoSessionePdi(difettoSessione: DetailedStoricoDifetto): Observable<DetailedStoricoDifetto>
  {
    return this._updateEntity(storicoDifettiStore, difettoSessione as StoricoDifettoDB, difettoSessione.GUID).pipe(
      map(m => (
      {
        ...m,
        ProdottoDescrizione: difettoSessione.ProdottoDescrizione,
        GruppoDescrizione: difettoSessione.GruppoDescrizione,
        ParticolareDescrizione: difettoSessione.ParticolareDescrizione,
        ComponenteDescrizione: difettoSessione.ComponenteDescrizione,
        DifettoDescrizione: difettoSessione.DifettoDescrizione,
        PosizioneDifettoDescrizione: difettoSessione.PosizioneDifettoDescrizione,
        PesoDifettoOldCodice: difettoSessione.PesoDifettoOldCodice,
        PesoDifettoNewCodice: difettoSessione.PesoDifettoNewCodice,
        ResponsabilitaString: difettoSessione.ResponsabilitaString,
        OriginiString: difettoSessione.OriginiString
      }) as DetailedStoricoDifetto)
    );
  }
  
  getVeicoloByGUID(GUID: string){
      return this.localDBServ.getByKey<VeicoloDB>(veicoliStore, GUID);
  }

  getVeicoloByGUIDSessione(GUIDSessione: string){
    return this.getSessioneByGUID(GUIDSessione)
      .pipe(
        switchMap(sessione => this.getVeicoloByGUID(sessione.VeicoloGUID))
      );
  }

  getVeicoloByVAN(VAN: string){
      return this.localDBServ.getByIndex<VeicoloDB>(veicoliStore, 'VAN', VAN);
  }

  getVeicoloByCategoria(categoria: string){
      return this.localDBServ.getByIndex<VeicoloDB>(veicoliStore, 'Categoria', categoria);
  }

  getVeicoliByVANOrCategoria(VAN: string, categoria: string){
    return this.localDBServ.getAll<VeicoloDB>(veicoliStore)
      .pipe(
        map(veicoli => this._filterVeicoliByVANOrCategoria(veicoli, VAN, categoria)),
      );
  }

  getVeicoliByGUIDs(GUIDs: string[]): Observable<VeicoloDB[]>{
    return this.localDBServ.bulkGet<VeicoloDB>(veicoliStore, GUIDs)
      .pipe(
        // di default se un elemento non viene trovato, viene
        // restituito "undefined" per tale elemento
        map(veicoli => veicoli.filter(m => !!m))
      );
  }

  getAllVeicoli(){
    return this.localDBServ.getAll<VeicoloDB>(veicoliStore);
  }

  getFullVeicoliByGUIDs(GUIDs: string[]): Observable<FullVeicoloDB[]> {
    return this.getVeicoliByGUIDs(GUIDs)
        .pipe(
            switchMap(veicoli => forkJoin([...veicoli.map(m => this._getFullVeicoloDB(m))])),
            defaultIfEmpty([])
        );
  }

  getFullVeicoloByGUID(GUID: string): Observable<FullVeicoloDB> {
    return this.getFullVeicoliByGUIDs([ GUID ]).pipe(
      map(m => m.length >= 1 ? m[0] : null)
    );
  }

  addVeicolo(toAdd: VeicoloDB){
    toAdd.GUID = this._getVeicoloGUID(toAdd.VAN, toAdd.Categoria);

    return this._addEntity<VeicoloDB>(veicoliStore, toAdd);
  }

  addFullVeicoli(veicoli: FullVeicoloDB[]){
    return from(veicoli)
      .pipe(
        concatMap(veicolo => this.addFullVeicolo(veicolo)),
        // affinchè l'emissione venga eseguita anche
        // se non c'è alcun veicolo da eliminare
        defaultIfEmpty(0),
        map(_ => veicoli.length),
        takeLast(1)
    );
  }

  addFullVeicolo(veicolo: FullVeicoloDB) {
    return of(veicolo)
      .pipe(
        map(veicolo => ({
            veicolo: this._getVeicoloFromFullVeicolo(veicolo),
            sessioni: veicolo.Sessioni.map(s => this._getSessioneFromFullSessione(s)),
            difetti: this.utilsServ.selectMany(veicolo.Sessioni, s => s.Difetti)
          })),
        switchMap(toAdd => forkJoin([
          this.localDBServ.bulkAdd<VeicoloDB[]>(veicoliStore, [toAdd.veicolo]),
          this.localDBServ.bulkAdd<SessioneDB[]>(sessioniStore, toAdd.sessioni),
          this.localDBServ.bulkAdd<StoricoDifettoDB[]>(storicoDifettiStore, toAdd.difetti)
        ])),
        map(_ => null),
        take(1)
      );
  }

  // questo metodo restituisce sia veicolo che sessione
  // poichè se il GUID del veicolo è cambiato, anche il riferimento
  // del veicolo sulle sessioni (di qualsiasi tipo) va aggiornato di conseguenza..
  updateVeicolo(toUpdate: VeicoloDB, currentSessioneGUID: string){
    return this._isGUIDVeicoloUnchanged(toUpdate)
      .pipe(
        switchMap(isGUIDVeicoloUnchanged => {

          if (isGUIDVeicoloUnchanged){
            return this._updateEntity(veicoliStore, toUpdate, toUpdate.GUID)
              .pipe(
                switchMap(veicolo => forkJoin([of(veicolo), this.getSessioneByGUID(currentSessioneGUID)])),
                map(result => ({ veicolo: result[0], sessione: result[1] }))
              );
          }

          // altrimenti, il GUID è cambiato:
          // tengo traccia del vecchio GUID..
          const oldGUID = toUpdate.GUID;
          // riassegno il GUID sul veicolo..
          toUpdate.GUID = this._getVeicoloGUID(toUpdate.VAN, toUpdate.Categoria);

          // estraggo tutte le sessioni riferite al vecchio GUID veicolo
          // (sia di AUDIT che di PDI)
          return this.getSessioniVeicolo(oldGUID)
            .pipe(
              // aggiorno il GUID del veicolo puntato su ciascuna sessione
              map(sessioni => sessioni.map(s => ({ ...s, VeicoloGUID: toUpdate.GUID} as SessioneDB))),
              // eseguo quindi l'aggiornamento in parallelo del veicolo e di tutte le sessioni ad esso puntate,
              // che vedranno quindi aggiornato il campo VeicoloGUID
              switchMap(sessioni => forkJoin([
                this._updateEntity(veicoliStore, toUpdate, oldGUID), 
                ...sessioni.map(s => this._updateEntity(sessioniStore, s, s.GUID))])),
              // coerentemente con il caso più semplice all'inizio del metodo, restituisco veicolo e sessione corrente aggiornate
              // (ma contestualmente anche le altre eventuali sessioni legate al veicolo sono state aggiornate...)
              map(result => ({
                veicolo: result[0] as VeicoloDB,
                sessione: (result.slice(1) as SessioneDB[]).find(m => m.GUID === currentSessioneGUID) as SessioneDB
              }))
            );
        })
      );
  }

  deleteFullVeicoliFromGUIDs(GUIDs: string[]){
    return this.getFullVeicoliByGUIDs(GUIDs)
      .pipe(
        switchMap(veicoli => this.deleteFullVeicoli(veicoli)),
        take(1)
      );
  }

  deleteFullVeicoli(veicoli: FullVeicoloDB[]){
    return from(veicoli)
      .pipe(
        concatMap(veicolo => this.deleteFullVeicolo(veicolo)),
        // affinchè l'emissione venga eseguita anche
        // se non c'è alcun veicolo da eliminare
        defaultIfEmpty(0),
        map(_ => veicoli.length),
        takeLast(1)
    );
  }

  deleteFullVeicolo(veicolo: FullVeicoloDB) {
    return of(veicolo)
      .pipe(
        map(veicolo => ({
          veicolo: this._getVeicoloFromFullVeicolo(veicolo),
          sessioni: veicolo.Sessioni.map(s => this._getSessioneFromFullSessione(s)),
          difetti: this.utilsServ.selectMany(veicolo.Sessioni, s => s.Difetti)
        })),
        switchMap(toDelete => forkJoin([
          this._deleteElementsByGUIDs(veicoliStore, [toDelete.veicolo.GUID]),
          this._deleteElementsByGUIDs(sessioniStore, toDelete.sessioni.map(s => s.GUID)),
          this._deleteElementsByGUIDs(storicoDifettiStore, toDelete.difetti.map(s => s.GUID))
        ])),
        map(_ => null),
        take(1)
      );
  }

  getAllSessioni(){
    return this.localDBServ.getAll<SessioneDB>(sessioniStore);
  }

  getSessioniVeicolo(vanGUID: string, tipoSessione: TipoSessione | undefined = null){
      const toReturn = this.localDBServ.getAllByIndex<SessioneDB>(sessioniStore, 'VeicoloGUID', IDBKeyRange.only(vanGUID));

      if (!tipoSessione){
        return toReturn;
      } else {
        return toReturn.pipe(
          map(m => m.filter(s => s.Tipo === tipoSessione))
        );
      }
  }

  getSessioneByGUID(GUID: string): Observable<SessioneDB> {
    return this.localDBServ.getByKey<SessioneDB>(sessioniStore, GUID);
  }

  addSessione(toAdd: SessioneDB){
      toAdd.GUID = uuidv4();

      return this._addEntity<SessioneDB>(sessioniStore, toAdd);
  }

  addFullSessione(sessione: FullSessioneDB): Observable<any> {
    sessione.GUID = uuidv4();

    return of(sessione)
      .pipe(
        // dopo aver assegnato il GUID alla nuova sessione, inizializzo
        // i difetti assegnando a ciascuno un GUID e legandoli al GUID della sessione
        map(s => ({
          ...s,
          Difetti: this._initDifettiNewSessione(s.Difetti, s.GUID)
        } as FullSessioneDB)),
        map(s => ({
          sessione: this._getSessioneFromFullSessione(s),
          difetti: s.Difetti
        })),
        switchMap(toAdd => forkJoin([
          this._addEntity<SessioneDB>(sessioniStore, toAdd.sessione),
          toAdd.difetti.map(difetto => this._addEntity<StoricoDifettoDB>(storicoDifettiStore, difetto))
        ])),
        map(_ => null),
        take(1)
      );
  }

  updateSessione(toUpdate: SessioneDB){
    return this._getEntityOrNull(sessioniStore, toUpdate.GUID)
      .pipe(
        map(sessione => {
          if (!sessione){
            throw Error('Impossibile trovare la sessione con GUID = ' + toUpdate.GUID);
          }

          const notNullSessione = sessione as SessioneDB;

          // non è ammessa la modifica di determinati dati..
          if (notNullSessione.GUID !== toUpdate.GUID){
            throw Error('La sessione estratta ha GUID diverso rispetto a quella passata al metodo');
          } else if (notNullSessione.VeicoloGUID !== toUpdate.VeicoloGUID){
            throw Error('La sessione estratta è riferita ad un veicolo diverso rispetto a passata al metodo');
          } else if (notNullSessione.Tipo !== toUpdate.Tipo) {
            throw Error('La sessione estratta è di tipo diverso rispetto a quella passata al metodo');
          }

          return toUpdate;
        }),
        switchMap(sessione => this._updateEntity<SessioneDB>(sessioniStore, sessione, sessione.GUID))
      );
  }

  clear(storeType: Type<any>) {
    switch(storeType){
      case ProdottoDB:
        return this.localDBServ.clear(prodottiStore);
      case GruppoDB:
        return this.localDBServ.clear(gruppiStore);
      case ParticolareDB:
        return this.localDBServ.clear(particolariStore);
      case ComponenteDB:
        return this.localDBServ.clear(componentiStore);
      case AnomaliaDB:
        return this.localDBServ.clear(anomalieStore);
      case PosizioneAnomaliaDB:
        return this.localDBServ.clear(posAnomalieStore);
      case PesoDifettoDB:
        return this.localDBServ.clear(pesiDifettoStore);
      case ResponsabilitaDB:
        return this.localDBServ.clear(responsabilitaStore);
      case GammaDB:
        return this.localDBServ.clear(gammeStore);
      case GammaSpecificaDB:
        return this.localDBServ.clear(gammeSpecificheStore);
      case ChecklistDB:
        return this.localDBServ.clear(checklistsStore);
      case ElementoChecklistDB:
        return this.localDBServ.clear(elementiChecklistsStore);
      default:
        throw new Error('Unexpected store type: ' + storeType.toString());
    }
  }

  bulkAdd<T>(storeType: Type<any>, elements: T[]){
    switch(storeType){
      case ProdottoDB:
        return this.localDBServ.bulkAdd(prodottiStore, elements);
      case GruppoDB:
        return this.localDBServ.bulkAdd(gruppiStore, elements);
      case ParticolareDB:
        return this.localDBServ.bulkAdd(particolariStore, elements);
      case ComponenteDB:
        return this.localDBServ.bulkAdd(componentiStore, elements);
      case AnomaliaDB:
        return this.localDBServ.bulkAdd(anomalieStore, elements);
      case PosizioneAnomaliaDB:
        return this.localDBServ.bulkAdd(posAnomalieStore, elements);
      case PesoDifettoDB:
        return this.localDBServ.bulkAdd(pesiDifettoStore, elements);
      case ResponsabilitaDB:
        return this.localDBServ.bulkAdd(responsabilitaStore, elements);
      case GammaDB:
        return this.localDBServ.bulkAdd(gammeStore, elements);
      case GammaSpecificaDB:
        return this.localDBServ.bulkAdd(gammeSpecificheStore, elements);
      case ChecklistDB:
        return this.localDBServ.bulkAdd(checklistsStore, elements);
      case ElementoChecklistDB:
        return this.localDBServ.bulkAdd(elementiChecklistsStore, elements);
      default:
        throw new Error('Unexpected store type: ' + storeType.toString());
    }
  }

  private _initDifettiNewSessione(difetti: StoricoDifettoDB[], GUIDSessione: string): StoricoDifettoDB[] {
    return difetti.map(difetto => ({
      ...difetto,
      SessioneGUID: GUIDSessione,
      GUID: uuidv4()
    } as StoricoDifettoDB));
  }

  private _isGUIDVeicoloUnchanged(toUpdate: VeicoloDB) {
    const oldGUID = toUpdate.GUID;
    const newGUID = this._getVeicoloGUID(toUpdate.VAN, toUpdate.Categoria);

    if (oldGUID === newGUID){
      return of(true);
    }

    return this.getVeicoloByGUID(newGUID)
      .pipe(
        switchMap(existingVeicolo => {
          if (!!existingVeicolo){
            return throwError('esiste già un veicolo con lo stesso VAN e/o la stessa categoria');
          }

          return of(false);
        })
      );
  }

  // N.B: da tenere allineato con metodo "GetCodiceDifetto" in ModelUtils del backend
  private _getCodiceDifetto(difetto: StoricoDifettoDB){
    let toReturn = '';

    if (!difetto.ProdottoCodice){
        throw new Error('Il difetto non è associato ad alcun prodotto');
    }

    toReturn += difetto.ProdottoCodice;

    if (!difetto.GruppoCodice){
        throw new Error('Il difetto non è associato ad alcun gruppo');
    }

    toReturn += difetto.GruppoCodice.substring(2);

    toReturn += (!!difetto.ParticolareCodice)
        ? difetto.ParticolareCodice.substring(4)
        : this._DIFETTO_NO_PARTICOLARE_AFFIX;

    toReturn += (!!difetto.ComponenteCodice)
        ? difetto.ComponenteCodice.substring(6)
        : this._DIFETTO_NO_COMPONENTE_AFFIX;

    if (!difetto.DifettoCodice){
        throw new Error('Il difetto non è associato ad alcuna anomalia');
    }

    toReturn += difetto.DifettoCodice;

    return toReturn;
  }

  private _getSessioneFromFullSessione(sessione: FullSessioneDB): SessioneDB {
    return { ...sessione, Difetti: undefined } as SessioneDB;
  }

  private _getVeicoloFromFullVeicolo(veicolo: FullVeicoloDB): VeicoloDB {
    return { ...veicolo, Sessioni: undefined } as VeicoloDB;
  }

  private _filterVeicoliByVANOrCategoria(allVeicoli: VeicoloDB[], VAN: string, categoria: string): VeicoloDB[]{
    // Implementa ricerca appoggiandosi agli indici dell'indexed DB, ma non consente ricerche tramite like.
    // Visto che l'implementazione attuale, che estrae comunque tutti i veicoli ha delle buone performances
    // si utilizza quest'ultima, che appunto gestisce in-memory anche le ricerche tramite like (includes)
    /*
    if (!!VAN){
      return this.localDBServ.getAllByIndex<VeicoloDB>(veicoliStore, 'VAN', IDBKeyRange.only(VAN));
    } else if (!!categoria){
      return this.localDBServ.getAllByIndex<VeicoloDB>(veicoliStore, 'Categoria', IDBKeyRange.only(categoria));
    } else {
      return this.localDBServ.getAll<VeicoloDB>(veicoliStore);
    } */

    return allVeicoli
      // se il campo di ricerca "VAN" è stato valorizzato,
      // considero i soli veicoli dotati di VAN che includa in tutto o in parte il VAN richiesto
      .filter(m => (!!VAN) ? !!m.VAN && m.VAN.toUpperCase().includes(VAN.toUpperCase()) : true)
      // se il campo di ricerca "categoria" è stato valorizzato,
      // considero i soli veicoli dotati di categoria che includa in tutto o in parte la categoria richiesta
      .filter(m => (!!categoria) ? !!m.Categoria && m.Categoria.toUpperCase().includes(categoria.toUpperCase()) : true);
  }

  private _getFullVeicoloDB(veicolo: VeicoloDB):Observable<FullVeicoloDB> {
    return this.getSessioniVeicolo(veicolo.GUID)
        .pipe(
            switchMap(sessioni => forkJoin([...sessioni.map(m => this._getFullSessioneDB(m))])),
            defaultIfEmpty([]),
            map(fullSessioni => ({ ...veicolo, Sessioni: fullSessioni}))
        );
  }

  private _getFullChecklistDB(checklist: ChecklistDB): Observable<FullChecklistDB> {
    return this.getElementiChecklists(true, checklist.Id).pipe(
      map(m => ({ ...checklist, Elementi: m } as FullChecklistDB))
    );
  }

  private _getFullSessioneDB(sessione: SessioneDB) : Observable<FullSessioneDB> {
    return this.getDifettiSessione(sessione.GUID)
        .pipe(
            map(difetti => ({ ...sessione, Difetti: difetti}))
        );
  }

  private _getDetailedDifetto(difetto: StoricoDifettoDB): Observable<DetailedStoricoDifetto> {
    const currentLanguageCode = this.translServ.getLanguage().code;

    return forkJoin([
      this.getProdottoByCodice(difetto.ProdottoCodice),
      this.getGruppoByCodice(difetto.GruppoCodice),
      this.getParticolareByCodice(difetto.ParticolareCodice),
      this.getComponenteByCodice(difetto.ComponenteCodice),
      this.getAnomaliaByCodice(difetto.DifettoCodice),
      this.getPosAnomaliaByCodice(difetto.PosizioneDifettoCodice),
      this.getPesoDifettoById(difetto.PesoDifettoId),
      this.getResponsabilita(false),
    ]).pipe(
      map(m => (
      {
        ...difetto,
        ProdottoDescrizione: this.utilsServ.getTraduzione(m[0] as ProdottoDB, currentLanguageCode),
        GruppoDescrizione: this.utilsServ.getTraduzione(m[1] as GruppoDB, currentLanguageCode),
        ParticolareDescrizione: this.utilsServ.getTraduzione(m[2] as ParticolareDB, currentLanguageCode),
        ComponenteDescrizione: this.utilsServ.getTraduzione(m[3] as ComponenteDB, currentLanguageCode),
        DifettoDescrizione: this.utilsServ.getTraduzione(m[4] as AnomaliaDB, currentLanguageCode),
        PosizioneDifettoDescrizione: this.utilsServ.getTraduzione(m[5] as PosizioneAnomaliaDB, currentLanguageCode),
        PesoDifettoOldCodice: (m[6] as PesoDifettoDB).CodiceOld,
        PesoDifettoNewCodice: (m[6] as PesoDifettoDB).CodiceNew,
        ResponsabilitaString: this._getResponsabilitaString(m[7] as ResponsabilitaDB[], difetto.ResponsabilitaIds),
        OriginiString: this._getResponsabilitaString(m[7] as ResponsabilitaDB[], difetto.OriginiIds),
      } as DetailedStoricoDifetto))
    );
  }

  private _getResponsabilitaString(responsabilita: ResponsabilitaDB[], selected: number[]){
    const toReturn = responsabilita
    .filter(m => !!selected && selected.includes(m.Id))
    .map(m => m.Nome)
    .reduce((prev, val) => val + ', ' + prev, '')
    .trim();

    return toReturn.replace(/^\,+|\,+$/g, '');
  }

  private _deleteElementsByGUIDs(storeName: string, GUIDs: string[]){
    return of(GUIDs)
    .pipe(
      switchMap(guids => forkJoin([...guids.map(m => this.localDBServ.deleteByKey(storeName, IDBKeyRange.only(m)))])),
      defaultIfEmpty([])
    );
  }

  private _getVeicoloGUID(van: string, categoria: string) {
    // mantenimento retrocompatibilità con audit PDI
    if (!!van && !categoria){
      return van;
    } else if (!van && !!categoria) {
      return 'NO_VAN_CAT_[' + categoria + ']';
    } else if (!!van && !!categoria) {
      return van + '_CAT_[' + categoria + ']';
    } else {
      throw new Error('VAN e Categoria non possono essere entrambi vuoti!');
    }
  }

  private _getEntityOrNull<T>(storeName: string, key: IDBValidKey){
    if (!key){
      return of(null);
    }

    return this.localDBServ.getByKey<T>(storeName, key);
  }

  private _updateEntity<T>(storeName: string, entity: T, key: Key){
    // NECESSARIO IN QUANTO METODO "updateByKey" E' BUGGATO:
    // vedi issue: https://github.com/assuncaocharles/ngx-indexed-db/issues/273
    return this.localDBServ.deleteByKey(storeName, key)
    .pipe(
      switchMap(_ => this._addEntity<T>(storeName, entity))
    );
  }

  private _addEntity<T>(storeName: string, entity: T): Observable<T>{
    // NECESSARIO IN QUANTO METODO "addItem" E' BUGGATO:
    // vedi issue: https://github.com/assuncaocharles/ngx-indexed-db/issues/263
      return this.localDBServ.bulkAdd<T[]>(storeName, [entity])
          .pipe(
              switchMap(result => this.localDBServ.getByKey<T>(storeName, result[0]))
          );
  }

  private _getEntitiesWithParent<T, TParentKey>(storeName: string, parentFieldName: string, parentKey: TParentKey): Observable<T[]> {
      return (parentKey !== null)
      ? this.localDBServ.getAllByIndex<T>(storeName, parentFieldName, IDBKeyRange.only(parentKey))
      : this.localDBServ.getAll<T>(storeName);
  }

  private _getEntitiesWithAbilitato<T>(storeName: string, abilitatoFieldName: string, onlyEnabled: boolean): Observable<T[]>{
      if (onlyEnabled){
          this.localDBServ.getAllByIndex<T>(storeName, abilitatoFieldName, IDBKeyRange.only(BooleanState.True));
      }

      return this.localDBServ.getAll<T>(storeName);
  }
}

