import { Injectable } from '@angular/core';
import { BehaviorSubject, Observable, merge, of, fromEvent, interval } from 'rxjs';
import { HttpClient, HttpHeaders } from '@angular/common/http';
import { mapTo, map, flatMap, catchError, tap, startWith, mergeMap } from 'rxjs/operators';
import { ExclusionHeaders } from '../_interceptors/exclusion-headers';
import { environment } from 'src/environments/environment';

export class PwaStatus {
  constructor(public isOnline: boolean, public ping?: number) { }
}

@Injectable()
export class PwaStatusService {

  private static _PING_URL = environment.apiUrl + '/pwa/ping';
  private static _PING_CYCLE_TIME = 5000;
  private _pwaStatus$ = new BehaviorSubject<PwaStatus>(null);

  constructor(private http: HttpClient) {
    this._getPwaStatusStream()
        .pipe(tap(updatedStatus => this._pwaStatus$.next(updatedStatus)))
        .subscribe();
   }

  public status$(){
    return this._pwaStatus$.asObservable();
  }

  private _getPwaStatusStream(){
    return merge(
        of(navigator.onLine).pipe(map(isOnline => new PwaStatus(isOnline))),
        fromEvent(window, 'online').pipe(mergeMap(_ => this._executePing())),
        fromEvent(window, 'offline').pipe(map(_ => new PwaStatus(false, -1))),
        this._getPingStream(PwaStatusService._PING_CYCLE_TIME));
  }

  private _getPingStream(cycleTime){
    return interval(cycleTime)
        .pipe(startWith(0),
            mergeMap(_ => this._executePing()));
  }

  private _executePing():Observable<PwaStatus>{
    // se il browser indica lo status di offline
    // vuol dire che non è nemmeno disponibile la connessione
    // ..inutile provare a pingare il server..
    if (!navigator.onLine){
        return of(new PwaStatus(false, -1));
    }

    console.log('pinging url: ' + PwaStatusService._PING_URL);
    const start = performance.now();

    let httpHeaders = new HttpHeaders();

    httpHeaders = httpHeaders.append(ExclusionHeaders.SPINNER_INTERCEPTOR, '_');
    httpHeaders = httpHeaders.append(ExclusionHeaders.ERROR_INTERCEPTOR, '_');

    return this.http.get(PwaStatusService._PING_URL, { headers: httpHeaders })
        .pipe(
            map(_ => +(performance.now() - start).toFixed(2)),
            map(ms => new PwaStatus(ms > 0, ms)),
            catchError(_ => of(new PwaStatus(false, -1)))
            );
  }
}
