import { Injectable, ElementRef } from '@angular/core';
import { GridComponent } from '@syncfusion/ej2-angular-grids';
import { IHasTranslations } from '../_interfaces/has-translations';
import { Elemento } from 'src/app/core/_models/common';

@Injectable()
export class UtilsService {

    constructor() { }

    public getTraduzione<T extends IHasTranslations>(element: T, currentLanguageCode: string, codePropertyForLocalePrefix: string = null): string {
        if (!element){
            return '';
        }

        const codePrefix = (!!codePropertyForLocalePrefix) ? element[codePropertyForLocalePrefix].toString() + ' - ' : '';

        return !!(element.Traduzioni.find(s => s.Code === currentLanguageCode))
        ? codePrefix + element.Traduzioni.find(s => s.Code === currentLanguageCode).Descrizione
        : codePrefix + element.Descrizione;
    }

    public clone<T>(obj: T): T {
        let copy;

        // Handle the 3 simple types, and null or undefined
        if (null == obj || 'object' !== typeof obj) {
            return obj;
        }

        // Handle Date
        if (obj instanceof Date) {
            copy = new Date();
            copy.setTime(obj.getTime());
            return copy;
        }

        // Handle Array
        if (obj instanceof Array) {
            copy = [];
            for (let i = 0, len = obj.length; i < len; i++) {
                copy[i] = this.clone(obj[i]);
            }
            return copy;
        }

        // Handle Object
        if (obj instanceof Object) {
            copy = {};
            for (const attr in obj) {
                if (obj.hasOwnProperty(attr)){
                    copy[attr] = this.clone(obj[attr]);
                }
            }
            return copy;
        }

        throw new Error('Unable to copy obj! Its type isn\'t supported.');
    }

    public getEnumDatasource<T>(numberKeysEnum: T):Elemento[] {     
        return Object.keys(numberKeysEnum)
        .filter(key => !isNaN(+key))
        .map(key => ({ Id: +key, Nome: numberKeysEnum[key]}));
    }

    public getTodayString() {
        const date = new Date();

        const year = String(date.getFullYear());

        let month = String(date.getMonth() + 1);

        if (month.length === 1) {
            month = '0' + month;
        }

        let day = String(date.getDate());
        if (day.length === 1) {
            day = '0' + day;
        }

        return year + '_' + month + '_' + day;
    }

    //converte dataURL immagine in BLOB per poter fare post
    public dataURItoBlob(dataURI) {
        // convert base64/URLEncoded data component to raw binary data held in a string
        let byteString;
        if (dataURI.split(',')[0].indexOf('base64') >= 0)
            byteString = atob(dataURI.split(',')[1]);
        else
            byteString = unescape(dataURI.split(',')[1]);

        // separate out the mime component
        const mimeString = dataURI.split(',')[0].split(':')[1].split(';')[0];

        // write the bytes of the string to a typed array
        const ia = new Uint8Array(byteString.length);
        for (let i = 0; i < byteString.length; i++) {
            ia[i] = byteString.charCodeAt(i);
        }

        return new Blob([ia], {type:mimeString});
      }

    public getSelectedRow<T>(args): T {
        const firstOuterGrid = this.getOuterGrid(args);

        if (!firstOuterGrid){
            return undefined;
        }

        // VECCHIA VERSIONE: la nuova consente di intercettare la riga corrente
        // anche quando si clicca sull'icona per espandere / collassare una certa riga
        // return firstOuterGrid.getRowInfo(evt.target).rowData as T;

        const currentRow = args.target.closest('.e-row');

        if (!currentRow){
            return undefined;
        }

        return firstOuterGrid.getRowObjectFromUID(currentRow.getAttribute('data-uid')).data as T;
    }

    public getOuterGrid(args) {
        const gridElement = args.target.closest('.e-grid');

        if (!gridElement) {
            return undefined;
        }

        return gridElement.ej2_instances[0];
    }

    public getContainedGrid(container: ElementRef, gridId: string){
        return container.nativeElement.querySelector('#' + gridId).ej2_instances[0];
    }

    // N.B.: affinchè la logica di expand/collapse funzioni correttamente, è necessario che il metodo
    // chiamante sia bindato all'evento click sulla griglia stessa utilizzando la seguente sintassi:
    // "this.grid.element.addEventListener('click', this.metodoChiamante.bind(this), true);"
    // -> il parametro "true" in fondo è FONDAMENTALE perchè evita il preventDefault sull'evento
    // di click sull'expand/collapse di una riga, in modo che oltre ad eseguire il metodo chiamante
    // sia eseguito anche il normale expand/collapse della riga stessa, cioè il comportamento di default
    public allowSingleExpandedRow(args){
        const grid = this.getOuterGrid(args);

        if (args.target.closest('.e-detailrowcollapse')) {
            grid.detailRowModule.collapseAll();

            const innerGridsElements = grid.element.getElementsByClassName('e-grid');

            for(let i = 0; i < innerGridsElements.length; i++){
                let innerGrid = innerGridsElements[i].ej2_instances[0];
                
                innerGrid.destroy();
            }

            // oltre a fare il collapse del dettaglio, lo RIMUOVO
            // anche dal DOM affinchè vengano liberate le rispettive risorse
            // TODO: anche con questa porzione di codice, la videata con la griglia
            // gerarchica continua a rallentarsi se viene usata massivamente..
            let detailRows = grid.element.getElementsByClassName('e-detailrow');
            while(detailRows.length > 0){
                detailRows[0].remove();
            }
        }
    }

    public isRowExpanderClick(args, gridIdPrefix?: string){
        if (!this.isCurrentGrid(args, gridIdPrefix)){
            return false;
        }

        return (args.target.closest('.e-detailrowcollapse'))
            || (args.target.closest('.e-detailrowexpand'));
    }

    private isCurrentGrid(args, gridIdPrefix){
        if (!gridIdPrefix){
            return true;
        }

        const grid = this.getOuterGrid(args);

        return grid.element.id.startsWith(gridIdPrefix);
    }

    public getExpandedRow<T>(args) {
        const grid = this.getOuterGrid(args);

        // se sto espandendo una riga..
        if (args.target.closest('.e-detailrowcollapse')) {

            // ..la restituisco
            return this.getSelectedRow<T>(args);
        } else if (args.target.closest('.e-detailrowexpand')) {
            // ..altrimenti vuol dire che sto collassando una riga,
            // nessuna riga è espansa e restituisco quindi undefined
            return undefined;
        }

        return undefined;
    }

    public expandDetailRow<T>(grid: GridComponent, condition: (T) => boolean) {

        for (const row of grid.getRowsObject()) {
            const rowData = row.data as T;

            if (condition(rowData)) {
                // per qualche ragione (approfondiremo in seguito) nella videata
                // di dettaglio di un CQ, gli eventi "dataBound" delle griglie figlio
                // (operations e phases) vengono chiamati 3 volte anzichè 1:
                // in entrambi i casi, il secondo evento va in eccezione in questa riga,
                // anche se poi l'espansione viene eseguita correttamente, verosimilmente
                // dalla prima o dalla terza chiamata dell'evento di cui sopra)
                try {
                     grid.detailRowModule.expand(row.index);
                }
                catch {
                    console.log('catch eccezione durante ri-espansione automatica riga dettaglio grid');
                }

                break;
            }
        }
    }

    public selectMany<TIn, TOut>(input: TIn[], selectListFn: (t: TIn) => TOut[]): TOut[] {
        return input.reduce((out, inx) => {
            out.push(...selectListFn(inx));
            return out;
        }, new Array<TOut>());
    }

    public sortFunc<TObject>(propFunc: (obj: TObject) => any,  direction: string = 'asc'){
        return (n1: TObject, n2: TObject) => {
            if (propFunc(n1) > propFunc(n2)) {
                return (direction === 'asc') ? 1 : -1;
            }

            if (propFunc(n1) < propFunc(n2)) {
                return (direction === 'asc') ? -1 : 1;
            }

            return 0;
        }
    }
}

// vecchia versione ricorsiva del metodo
// getOuterGrid non più necessaria, grazie a metodo 'closest'
/*const ej2InstancesSelector = 'ej2_instances';

if (!target) {
    return null;
} else if (target.classList && target.classList.contains('e-grid')){
    return target[ej2InstancesSelector][0];
} else {
    return this.getOuterGrid(target.parentElement);
} */
