import { Component, OnInit, ViewChild, EventEmitter, Input, ViewEncapsulation, OnDestroy, Output, OnChanges, ChangeDetectionStrategy } from '@angular/core';
import { GridsterComponent, GridsterConfig, GridType, DisplayGrid, GridsterItem, GridsterItemComponentInterface } from 'angular-gridster2';
import { CustomGridsterItem, DashboardConfig, DashboardConfigServiceBase, IDashboardConfigService } from 'src/app/core/_models/dashboard';
import { Subscription } from 'rxjs';
import { ActivatedRoute, Router } from '@angular/router';
import { DashboardItemComponent } from './dashboard-item/dashboard-item.component';

@Component({
  selector: 'app-dashboard',
  templateUrl: './dashboard.component.html',
  styleUrls: ['./dashboard.component.scss'],
  encapsulation: ViewEncapsulation.None
})
export class DashboardComponent implements OnInit, OnDestroy, OnChanges {

  dashboardItem = DashboardItemComponent;

  @ViewChild('gridster', {static: false})
  gridster: GridsterComponent

  // visto che il servizio in inpu non cambia (a meno che non cambi proprio la tipologia di dashboard..)
  // tramite questo campo di input sono in grado di segnalare dall'esterno al dashboard che si deve aggiornare
  @Input()
  toggleRefresh: boolean;

  dashboardConfig: DashboardConfig;
  gridsterConfig: GridsterConfig;

  private subscriptions = new Subscription();

  availableItemsFields = { text: 'title', value: 'type' };

  maximizedItem: CustomGridsterItem;
  beforeMaximizedScrollY: number;

  itemsOutput = {
    itemMaxSizeRequested: (item) => this.onMaxSizeItem(item),
    itemNormalSizeRequested: (item) => this.onNormalSizeItem(item),
    itemRemoveRequested: (item) => this.onRemoveItem(item)
  }

  constructor(private configService: DashboardConfigServiceBase) { }

  ngOnInit() {
      this._init();
   }

  ngOnDestroy() {
    this.subscriptions.unsubscribe();
  }

  ngOnChanges(changes) {
    if (changes.toggleRefresh.previousValue === undefined){
      return;
    }

    this._init();
  }

  private _init(){
    console.log('INIZIALIZZAZIONE DASHBOARD...');

    this.maximizedItem = null;

    this.dashboardConfig = this.configService.getConfig();
    
    this.gridsterConfig = {
      gridType: this.dashboardConfig.GridType ?? GridType.VerticalFixed,
      displayGrid: this.dashboardConfig.DisplayGrid ?? DisplayGrid.OnDragAndResize,
      minCols: this.dashboardConfig.MinCols ?? 10,
      maxCols: this.dashboardConfig.MaxCols ?? 10,
      minRows: this.dashboardConfig.MinRows ?? 10,
      maxRows: this.dashboardConfig.MaxRows ?? 40,
      fixedRowHeight: this.dashboardConfig.FixedRowHeight ?? 50,
      fixedColWidth: this.dashboardConfig.FixedColWidth ?? null,
      // fondamentale per fare in modo che i pannelli 
      // mantengano l'altezza anche in visualizzazione mobile
      keepFixedHeightInMobile:true,
      scrollSpeed: 10,
      scrollSensitivity: 0.001,
      pushItems: true,
      draggable: {
        enabled: true
      },
      resizable: {
        enabled: true
      },
      itemChangeCallback: this.onMoveItem.bind(this),
    };

    console.log('...COMPLETATA');
  }

  onMoveItem(item: GridsterItem, itemComponent: GridsterItemComponentInterface){
    this.configService.updateItems(this.dashboardConfig.Items);
  }

  onRemoveItem(item: CustomGridsterItem) {
    const toRemove = this.dashboardConfig.Items.filter(m => m.type === item.type);

    if (!toRemove){
      return;
    }

    this.dashboardConfig.Items = this.dashboardConfig.Items.filter(m => m.type !== item.type);

    this.configService.updateItems(this.dashboardConfig.Items);

    this.dashboardConfig.AvailableItems = this.configService.getAvailableItems();
  }

  onAddItem(toAdd) {
    const existingItem = this.dashboardConfig.Items.filter(m => m.type === toAdd.type);

    if (!existingItem){
      return;
    }

    // serve lo spread operator per evitare che coincidano le reference
    // degli oggetti nelle diverse liste di dashboardConfig
    this.dashboardConfig.Items.push({...toAdd });
    this.configService.updateItems(this.dashboardConfig.Items);
    this.dashboardConfig.AvailableItems = this.configService.getAvailableItems();
  }

  onResetItems() {
    // anche qui, si usa lo spread operator per la stessa ragione indicata sopra...
    this.dashboardConfig.Items = this.dashboardConfig.DefaultItems.map(s => {return {...s}});
    this.configService.updateItems(this.dashboardConfig.Items);
    this.dashboardConfig.AvailableItems = this.configService.getAvailableItems();
  }

  onMaxSizeItem(item: CustomGridsterItem){
    this.beforeMaximizedScrollY = this.gridster.el.scrollTop;
    this.scrollTo(this.gridster.el, 0,0);
    this.maximizedItem = item;
  }

  onNormalSizeItem(item: CustomGridsterItem) {
    this.maximizedItem = null;
    this.scrollTo(this.gridster.el, 0, this.beforeMaximizedScrollY);
  }

  scrollTo(el: any, x: number, y: number){
    try{
      this.gridster.el.scrollTo(x, y);
    } catch {
      this.gridster.el.scrollTop = 0;
    }
  }

  shouldMaximize(item: CustomGridsterItem){
    if (this.maximizedItem === item){
      return true;
    }

    return false;
  }
}
