import { Component, OnInit, Input, ViewChild, ViewEncapsulation, OnDestroy } from '@angular/core';
import { FileManagerComponent as EJ2FileManagerComponent, isFile } from '@syncfusion/ej2-angular-filemanager';
import { AuthenticationService } from 'src/app/core/_services/authentication.service';
import { environment } from 'src/environments/environment';
import { DialogComponent } from '@syncfusion/ej2-angular-popups';
import { ExtendedAjaxSettingsModel } from 'src/app/core/_models/extended-ajax-settings';
import * as printJS from 'print-js'
import { HttpClient } from '@angular/common/http';
import { map } from 'rxjs/operators';
import { Base64StringModel, GetBase64FullDocument } from 'src/app/core/_models/common';
import { Subscription } from 'rxjs';

enum FileType {
  Image = 0,
  PDF = 1
}

enum ImageTransform{
  RotateRight = 0,
  RotateLeft = 1,
  FlipVertically = 2,
  FlipHorizontally = 3
}

class FileTypeExtensionMapping {
  Type: FileType;
  Extensions: string[];
}

class FileInfo {
  Name: string;
  Type: FileType;
  RelativePath: string;
  Base64String?: string;
}

@Component({
  selector: 'app-file-manager',
  templateUrl: './file-manager.component.html',
  styleUrls: ['./file-manager.component.scss'],
})
export class FileManagerComponent implements OnInit, OnDestroy {

  public static readonly FILE_TYPE_IMAGE = 0;
  public static readonly FILE_TYPE_PDF = 1;

  private subscriptions = new Subscription();

  @ViewChild('fileManager', { static: false})
  public fileManager: EJ2FileManagerComponent;

  @ViewChild('FileViewerDialog', { static: false})
  public FileViewerDialog: DialogComponent;

  @Input()
  height = '600px';

  @Input()
  rootFolderName = 'Root folder';

  @Input()
  ajaxSettings: ExtendedAjaxSettingsModel;

  fileInfo: FileInfo;

  fileTypeExtMappings: Array<FileTypeExtensionMapping>;

  FileViewerDialogButtons = [
      { click: this.closeFileViewerDialog.bind(this), buttonModel: { content: 'Chiudi', isPrimary: 'true' } },
  ];

  constructor(private authService: AuthenticationService, private httpClient: HttpClient) {
    this.fileTypeExtMappings = [
      {Type: FileType.PDF, Extensions: ['PDF']}, 
      {Type: FileType.Image, Extensions: ['PNG', 'JPG', 'JPEG', 'BMP', 'TIF', 'TIFF', 'GIF']}
    ];
  }

  ngOnInit(){ }

  ngOnDestroy(){ 
    this.subscriptions.unsubscribe();
  }

  beforeSend(args: any) {
    // gestione autorizzazione
    args.ajaxSettings.beforeSend = (function (args) { 
      args.httpRequest.setRequestHeader('Authorization', 'Bearer ' + this.authService.currentUserToken);
    }).bind(this);
  }

  fileOpen(args) {
    args.cancel = true;

    this.fileInfo = this.getFileInfo(args.fileDetails);

    if (!this.fileInfo){
      return;
    }

    this.FileViewerDialog.show();

    this.subscriptions.add(this.getDocumentAsBase64String(this.fileManager.path, this.fileInfo.Name)
        .subscribe(file => {
          this.fileInfo.Base64String = file;
      }));
  }

  getDocumentAsBase64String(relativePath: string, fileName: string){
    const toPost = new GetBase64FullDocument();

    toPost.RelativePath = relativePath;
    toPost.FileName = fileName;

    return this.httpClient.post<Base64StringModel>(this.ajaxSettings.fullDocumentBase64Url, toPost)
      .pipe(map(result => result.Base64String));
  }

  getFileInfo(fileDetails):FileInfo {
    if (!isFile){
      return null;
    }

    const supportedFileType = this.getSupportedFileType(fileDetails.name);

    if (supportedFileType === null){
      return null;
    }

    const toReturn = new FileInfo();

    toReturn.Type = supportedFileType;
    toReturn.Name = fileDetails.name;
    toReturn.RelativePath = fileDetails.filterPath + fileDetails.name;

    return toReturn;
  }

  getSupportedFileType(fileName: string): FileType {
    const fileNameToUpper = fileName.toUpperCase();

    const toReturn = this.fileTypeExtMappings.filter(m => m.Extensions.filter(s => fileNameToUpper.endsWith(s)).length > 0);

    if (toReturn.length <= 0){
      return null;
    }

    return toReturn[0].Type;
  }

  isMatchingFileReadyForPreview(fileType: number){
    if (!this.fileInfo){
      return false;
    }

    if (!this.fileInfo.Base64String){
      return false;
    }

    const currentFileType = this.fileInfo.Type as number;

    return fileType === currentFileType;
  }

  // GESTIONE ROTAZIONE E FLIP VERTICALE/ORIZZONTALE IMMAGINE


  public onSnapshotRotation(clockwise: boolean){
    const transformType = clockwise
    ? ImageTransform.RotateRight
    : ImageTransform.RotateLeft;

    this._transformImage(transformType);
  }

  public onSnapshotFlip(horizontal: boolean){
    const transformType = horizontal
    ? ImageTransform.FlipHorizontally
    : ImageTransform.FlipVertically;

    this._transformImage(transformType);
  }

  private _transformImage(operation: ImageTransform) {
    // create an off-screen canvas
    const offScreenCanvas = document.createElement('canvas');
    const offScreenCanvasCtx = offScreenCanvas.getContext('2d');

    // create Image
    const img = new Image();

    // when the image is loaded..
    img.onload = () => {

      switch(operation)
      {
        case ImageTransform.RotateRight:
          offScreenCanvas.height = img.width;
          offScreenCanvas.width = img.height;
          offScreenCanvasCtx.rotate(90 * Math.PI / 180);
          offScreenCanvasCtx.translate(0, -offScreenCanvas.width);
          break;
        case ImageTransform.RotateLeft:
          offScreenCanvas.height = img.width;
          offScreenCanvas.width = img.height;
          offScreenCanvasCtx.rotate(-90 * Math.PI / 180);
          offScreenCanvasCtx.translate(-offScreenCanvas.height, 0);
          break;
        case ImageTransform.FlipHorizontally:
          offScreenCanvas.height = img.height;
          offScreenCanvas.width = img.width;
          offScreenCanvasCtx.translate(offScreenCanvas.width, 0);
          offScreenCanvasCtx.scale(-1, 1);
          break;
        case ImageTransform.FlipVertically:
          offScreenCanvas.height = img.height;
          offScreenCanvas.width = img.width;
          offScreenCanvasCtx.translate(0, offScreenCanvas.height);
          offScreenCanvasCtx.scale(1, -1);
          break;
        default:
        break;
      }

      offScreenCanvasCtx.drawImage(img, 0, 0);

      const mimeType = 'image/jpeg';
      const dataUrl = offScreenCanvas.toDataURL(mimeType);
      this.fileInfo.Base64String = dataUrl.replace('data:image/jpeg;base64,', '');
      offScreenCanvas.remove();

      // TODO SALVARE NUOVA IMMAGINE IN BACKEND
    }

    // chiamata che scatena callback "onload" sopra definita
    img.src = 'data:image/jpeg;base64,' + this.fileInfo.Base64String;
  }


  onPDFPrint(args){
    // print with print-js
    printJS({printable: this.fileInfo.Base64String, type: 'pdf', base64: true});
  }

  closeFileViewerDialog(){
    this.FileViewerDialog.hide();
  }

  refreshFiles(){
    this.fileManager.refreshFiles();
  }
}
