import { ChangeDetectorRef, Component, EventEmitter, HostListener, Input, OnInit, Output, ViewChild, ViewEncapsulation } from '@angular/core';
import { ImageDrawingComponent } from 'ngx-image-drawing';
import { WebcamImage, WebcamInitError, WebcamUtil } from 'ngx-webcam';
import { Observable, Subject } from 'rxjs';

enum WebcamImageTransform{
  RotateRight = 0,
  RotateLeft = 1,
  FlipVertically = 2,
  FlipHorizontally = 3
}
@Component({
  selector: 'app-webcam-standalone-viewer',
  templateUrl: './webcam-standalone-viewer.component.html',
  styleUrls: ['./webcam-standalone-viewer.component.scss'],
  encapsulation: ViewEncapsulation.None
})
export class WebcamStandaloneViewerComponent implements OnInit {

  @ViewChild('imgDrawing', {static: false})
  imgDrawing: ImageDrawingComponent;

  @Output()
  public snapshotConfirmed = new EventEmitter<string>();

  @Output()
  public dialogWidthPercentStringChanged = new EventEmitter<string>();

  // toggle webcam on/off
  public showWebcam = false;
  public allowCameraSwitch = true;
  public multipleWebcamsAvailable = false;

  public webcamImage:WebcamImage = null;
  public webcamError:WebcamInitError = null;

  private _dialogWidthPercent: number;

  @Input()
  public fixedHeight: number;
  @Input()
  public fixedWidth: number;

  public webcamHeight: number;
  public webcamWidth: number;

  public dialogWidthPercentString: string;

  public videoOptions: MediaTrackConstraints = {
     width: {ideal: 600},
     height: {ideal: 400},
     facingMode: { ideal: 'environment' }
  };

  public dialogHeight: number;

  // webcam snapshot trigger
  private trigger: Subject<void> = new Subject<void>();

  // switch to next / previous / specific webcam; true/false: forward/backwards, string: deviceId
  private nextWebcam: Subject<boolean|string> = new Subject<boolean|string>();

  constructor(private cdr: ChangeDetectorRef){}

  @HostListener('window:resize', ['$event'])
  onResize(event?: Event) {
    this._updateUI();
  }

  ngOnInit(): void {
    this._updateUI();

    WebcamUtil.getAvailableVideoInputs()
    .then((mediaDevices: MediaDeviceInfo[]) => {
      this.multipleWebcamsAvailable = mediaDevices && mediaDevices.length > 1;

      if (mediaDevices.length > 0){
        this.showWebcam = true;
      }
    });
  }

  private _updateUI(){
    this._updateDialogPercent();
    this._updateDialogWidthPercentString();
    this._updateWebcamSize();
  }

  private _updateDialogPercent(){
    const dialogWidth = Math.min(window.innerWidth * 0.9, 1280);

    this._dialogWidthPercent =  dialogWidth / window.innerWidth;
  }

  private _updateDialogWidthPercentString(){
    this.dialogWidthPercentString = (this._dialogWidthPercent * 100).toString() + '%';

    this.dialogWidthPercentStringChanged.emit(this.dialogWidthPercentString);
  }

  private _updateWebcamSize(){
    const size = window.innerWidth * this._dialogWidthPercent * 0.95;

    this.webcamWidth = this.fixedWidth ?? size;
    this.webcamHeight = this.fixedHeight ?? size;
  }

  handleInitError(error: WebcamInitError): void {
    this.showWebcam = false;
    this.webcamError = error;
  }

  onTakeSnapshot() {
    this.trigger.next();
  }

  onConfirmSnapshot(){
    this.imgDrawing.saveImage();
  }

  onImageDrawingSave(blob){
    const reader = new FileReader();
    reader.onloadend = () => {
        const base64data = reader.result as string;
        this.snapshotConfirmed.emit(base64data);
        this.webcamImage = null;

        // necessario per chiudere immediatamente
        // il dialog una volta completata l'acquisizione
        // dell'immagine: viceversa ci può mettere vari secondi
        // su dispositivi lenti...
        this.cdr.detectChanges();
    };

    reader.readAsDataURL(blob);
  }

  onDeleteSnapshot(){
    this.webcamImage = null;
  }

  handleSnapshot(webcamImage: WebcamImage): void {
    this.webcamImage = webcamImage;
  }

  public onSnapshotRotation(clockwise: boolean){
    const transformType = clockwise
    ? WebcamImageTransform.RotateRight
    : WebcamImageTransform.RotateLeft;

    this._transformImage(this.webcamImage, transformType, (rotatedImage) => {
      this.webcamImage = rotatedImage;
    });
  }

  public onSnapshotFlip(horizontal: boolean){
    const transformType = horizontal
    ? WebcamImageTransform.FlipHorizontally
    : WebcamImageTransform.FlipVertically;

    this._transformImage(this.webcamImage, transformType, (rotatedImage) => {
      this.webcamImage = rotatedImage;
    });
  }

  private _transformImage(webcamImage: WebcamImage, operation: WebcamImageTransform, callback) {
      // create an off-screen canvas
      const offScreenCanvas = document.createElement('canvas');
      const offScreenCanvasCtx = offScreenCanvas.getContext('2d');

      // create Image
      const img = new Image();

      // when the image is loaded..
      img.onload = () => {

        switch(operation)
        {
          case WebcamImageTransform.RotateRight:
            offScreenCanvas.height = img.width;
            offScreenCanvas.width = img.height;
            offScreenCanvasCtx.rotate(90 * Math.PI / 180);
            offScreenCanvasCtx.translate(0, -offScreenCanvas.width);
            break;
          case WebcamImageTransform.RotateLeft:
            offScreenCanvas.height = img.width;
            offScreenCanvas.width = img.height;
            offScreenCanvasCtx.rotate(-90 * Math.PI / 180);
            offScreenCanvasCtx.translate(-offScreenCanvas.height, 0);
            break;
          case WebcamImageTransform.FlipHorizontally:
            offScreenCanvas.height = img.height;
            offScreenCanvas.width = img.width;
            offScreenCanvasCtx.translate(offScreenCanvas.width, 0);
            offScreenCanvasCtx.scale(-1, 1);
            break;
          case WebcamImageTransform.FlipVertically:
            offScreenCanvas.height = img.height;
            offScreenCanvas.width = img.width;
            offScreenCanvasCtx.translate(0, offScreenCanvas.height);
            offScreenCanvasCtx.scale(1, -1);
            break;
          default:
          break;
        }

        offScreenCanvasCtx.drawImage(img, 0, 0);

        const mimeType = 'image/jpeg';
        const dataUrl = offScreenCanvas.toDataURL(mimeType);
        const imageData = offScreenCanvasCtx.getImageData(0, 0, offScreenCanvas.width, offScreenCanvas.height);

        offScreenCanvas.remove();

        callback(new WebcamImage(dataUrl, mimeType, imageData));
      }
      img.src = webcamImage.imageAsDataUrl;
  }

  showNextWebcam(): void {
    if (!this.multipleWebcamsAvailable){
      return;
    }
    // true => move forward through devices
    // false => move backwards through devices
    // string => move to device with given deviceId
    this.nextWebcam.next(true);
  }

  cameraWasSwitched(deviceId: string): void {
    console.log('active device: ' + deviceId);
  }

  get triggerObservable(): Observable<void> {
    return this.trigger.asObservable();
  }

  get nextWebcamObservable(): Observable<boolean|string> {
    return this.nextWebcam.asObservable();
  }

}
